/* Copyright 1999 VLSI Vision Ltd. Free for use under the Gnu Public License */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* next needed for definition of GetTime() */
#include <sys/time.h>
#include <time.h>

#include "vvl_defs.h"
#include "pport.h"

#include "packet.h"
#include "cmd.h"

#include "cpialink.h"

PPORT gPort[ MAX_LPT_PORTS+1 ];			  /* entry [0] not used !!! */
BOOL gPortInitialised[ MAX_LPT_PORTS+1 ]; /* entry [0] not used !!! */

PPORT *gCurrPort;


BOOL CLINK_Init( void ) {
	BYTE i;

	for( i = 0; i < MAX_LPT_PORTS; i++ ) {
		memset( &gPort[ i ], 0, sizeof( PPORT ) );
		gPortInitialised[ i ] = FALSE;
	}
	
	gCurrPort = NULL;

	return( TRUE );
}


void CLINK_Fini( void ) {
	BYTE i;

	for( i = 0; i < MAX_LPT_PORTS; i++ )
		if ( gPortInitialised[ i ] ) {
			ClosePPort( &gPort[ i ] );
			gPortInitialised[ i ] = FALSE;
		}
	gCurrPort = NULL;

	return;
}

BOOL CLINK_DetectPort( BYTE PortNumber, WORD *BaseAddr, BYTE *PortType ) {

 	if ( !DetectPortType( (byte)PortNumber, &gPort[PortNumber] ) ) {
        fprintf(stderr, "CLINK_DetectPort: LPT%d: not found\n", PortNumber );
		*BaseAddr = 0;
		*PortType = 0;
		return( FALSE );
	}
	*BaseAddr = gPort[PortNumber].BaseAddr;
	*PortType = gPort[PortNumber].PortType;

	return( TRUE );
}


BOOL CLINK_SelectPort( BYTE PortNumber ) {	/* will initialise port the first time its selected */

	if ( gPortInitialised[ PortNumber ] == FALSE ) {
	 	if ( !DetectPortType( PortNumber, &gPort[PortNumber] ) ) {
	        fprintf(stderr, "CLINK_SelectPort: LPT%d: not found\n", PortNumber );
			return( FALSE );
		}
		if ( !InitialisePPort( &gPort[PortNumber] ) ) {
	        fprintf(stderr, "CLINK_SelectPort: unable to initialise LPT%d\n", PortNumber );
			return( FALSE );
		}
	}

	gPortInitialised[ PortNumber ] = TRUE;
	gCurrPort = &gPort[ PortNumber ];

	return( TRUE );
}


BOOL CLINK_GetSelectedPort( BYTE *PortNumber ) {

	if ( gCurrPort == NULL ) return( FALSE );

	*PortNumber = gCurrPort->PortNumber;

	return( TRUE );
}


BOOL CLINK_SendWakeup( void ) {
	BYTE VPP[8] = { VPP_PREAMBLE0, 
					VPP_PREAMBLE1,
					VPP_PREAMBLE2,
					VPP_PREAMBLE3,
					VPP_PREAMBLE4,
					VPP_PREAMBLE5,
					VPP_MANUAL_CMD,
					VPP_TERMINATOR };
	if ( gCurrPort == NULL ) return( FALSE );
	SendVPP( gCurrPort, 8, &VPP[0] );
	return( TRUE );
}


BOOL CLINK_Get1284DeviceID( char *DeviceID, WORD MaxLength ) { 	/* use small MaxLength to 		*/ 
																/* check if camera alive		*/
	int rtn_int;
 	char nibble_string[ MAX_1284_DEVICE_ID ];
 
	if ( gCurrPort == NULL ) return( FALSE );

 	if ( !Get1284DeviceID( nibble_string ) ) return( FALSE );

   	strncpy( DeviceID, nibble_string, MaxLength );

	return( TRUE );
}


BOOL CLINK_ForceWatchdogReset( void ) {
	DWORD start;
	BYTE buff[4];

	if ( !SetTransferMode( gCurrPort, (REQUEST_DEVICE_ID | NIBBLE_TRANSFER) ) )
		 return( FALSE );

	if ( !NibbleReadBuffer( gCurrPort, (BYTE __huge *)buff, 4 ) ) 
		return( FALSE );

	/* now wait long enough to guarantee the CPIA Watchdog Timer will 	*/
	/* have timed out and reset the board. 							 	*/
    start = GetTime();                                        
	while( (((GetTime() - start)*1000)/(DWORD)CLOCKS_PER_SEC) < 1500); /* ie wait 1500mS */

	EndTransferMode( gCurrPort );

	return( TRUE );
}


BOOL CLINK_GetPortInfo( WORD *PortAddr, BYTE *PortType, WORD *Timeout_mS ) {
	
	if ( gCurrPort == NULL ) return( FALSE );
	
	if ( !GetTimeout( gCurrPort, Timeout_mS ) ) return( FALSE );

	*PortAddr = gCurrPort->BaseAddr;
	*PortType = gCurrPort->PortType;

//	fprintf(stderr, "GetPortInfo: Whileout = %lu\n", gCurrPort->Whileout );

	return( TRUE );
}


BOOL CLINK_SetPortInfo( BYTE PortType, WORD Timeout_mS ) {

	if ( gCurrPort == NULL ) return( FALSE );

	if ( !SetTimeout( gCurrPort, Timeout_mS ) ) return( FALSE );
	gCurrPort->PortType = PortType;

	return( TRUE );
}


int Get1284DeviceID( char __huge *DestString ) {
	WORD tmp;
	BYTE  length[20];
	
	DestString[0] = '\0';

	if ( !SetTransferMode( gCurrPort, (REQUEST_DEVICE_ID | NIBBLE_TRANSFER) ) )
		 return( FALSE );

	if ( !NibbleReadBuffer( gCurrPort, (BYTE  __huge *)length, 2 ) ) {
//		fprintf(stderr, "Get1284DeviceID:  NibbleReadBuffer returned an error\n" );
		EndTransferMode( gCurrPort );
		DestString[0] = '\0';
		return( FALSE );
	}

	tmp = length[0]*256 + length[1];

	if ( tmp < 3 ) {
		EndTransferMode( gCurrPort );
		return( FALSE );
	}
	tmp = tmp - 2;			/* reduce count to remove length BYTE s */
	if ( tmp >= MAX_1284_DEVICE_ID ) tmp = MAX_1284_DEVICE_ID - 1;

	if ( !NibbleReadBuffer( gCurrPort, (BYTE  __huge *)DestString, tmp ) ) {
		DestString[0] = '\0';
		EndTransferMode( gCurrPort );
		return( FALSE );
	}
	DestString[ tmp ] = '\0';
	
	if ( !EndTransferMode( gCurrPort ) ) return( FALSE );

 	return( TRUE );
}


BOOL CLINK_TransferMsg( BYTE  Module, BYTE  Proc, CMD_DATA *CmdData,
	BOOL DataReadFlag, WORD DataBytes, DATA_BUFFER *DataBuffer )
{
WORD gDataPacketsPending;
BYTE  Cmd[PACKET_LEN];
BYTE  __huge *DataPtr;

if ( gCurrPort == NULL ) return( FALSE );

Cmd[ CMD_TYPE ] = CAMERA_REQUEST;
if ( DataReadFlag ) Cmd[ CMD_TYPE ] |= READ_DATA_REQUEST;
	
Cmd[ CMD_DEST_ADDR ] =\
	(BYTE)((Module & MODULE_ADDR_BITS) | (Proc & PROC_ADDR_BITS));

if ( CmdData != NULL )
	{
	Cmd[ CMD_D0 ] = CmdData->Raw[0];
	Cmd[ CMD_D1 ] = CmdData->Raw[1];
	Cmd[ CMD_D2 ] = CmdData->Raw[2];
	Cmd[ CMD_D3 ] = CmdData->Raw[3];
	}
else
	{
	Cmd[ CMD_D0 ] = 0; Cmd[ CMD_D1 ] = 0;
	Cmd[ CMD_D2 ] = 0; Cmd[ CMD_D3 ] = 0;
	}

Cmd[ CMD_DATA_BYTES_LO ] = (BYTE)(DataBytes&0xFF);
Cmd[ CMD_DATA_BYTES_HI ] = (BYTE)(DataBytes>>8);

DataPtr = (BYTE  __huge *)DataBuffer;

if( !SetTransferMode( gCurrPort, ECP_TRANSFER ) )
	{
	fprintf(stderr, "TransferMsg: SetTransferMode failed\n" );
	goto TransferError;
	}
if( gCurrPort->PortType & ECP_PORT_TYPE )
	{
	if ( !ECPWriteBuffer(\
	gCurrPort, (BYTE  __huge *)Cmd, (DWORD)PACKET_LEN ) )
		{
		fprintf(stderr, "TransferMsg: ECPWriteBuffer failed\n" );
		EndTransferMode( gCurrPort );
		goto TransferError;
		}
	}
else
	{
	if ( !SimECPWriteBuffer(\
	gCurrPort, (BYTE  __huge *)Cmd, (DWORD)PACKET_LEN ) )
		{
		fprintf(stderr, "TransferMsg: SimECPWriteBuffer failed\n" );
		EndTransferMode( gCurrPort );
		goto TransferError;
		}
	}
if( !EndTransferMode( gCurrPort ) )
	{
	fprintf(stderr, "TransferMsg: EndTransferMode failed\n" );
	goto TransferError;
	}

if ( DataBytes != 0 )
	{
	gDataPacketsPending = DataBytes / PACKET_LEN;

	while( gDataPacketsPending )
		{
		if ( !DataReadFlag )
			{
			/* write data packet to camera */	    
			if ( !SetTransferMode( gCurrPort, ECP_TRANSFER ) )
				{
				fprintf(stderr, "TransferMsg: SetTransferMode failed\n" );
				fprintf(stderr, "TransferMsg: gDataPacketsPending = %u\n", gDataPacketsPending );
				goto TransferError;
				}
			if ( gCurrPort->PortType & ECP_PORT_TYPE )
				{
				if ( !ECPWriteBuffer(\
				gCurrPort, DataPtr, (DWORD)PACKET_LEN ) )
					{
					fprintf(stderr, "TransferMsg: ECPWriteBuffer failed\n" );
					goto TransferError;
					}
				}
			else
				{
				if ( !SimECPWriteBuffer(\
				gCurrPort, DataPtr, (DWORD)PACKET_LEN ) )
					{
					fprintf(stderr, "TransferMsg: SimECPWriteBuffer failed\n" );
					goto TransferError;
					}
				}
			if( !EndTransferMode( gCurrPort ) )
				{
				fprintf(stderr, "TransferMsg: EndTransferMode failed\n" );
				goto TransferError;
				}
			}
		else /* if DataReadFlag set */
 			{
			/* read data packet from camera */	 
 			if ( gCurrPort->PortType & BIDIR_PORT_TYPE )
 				{
				if ( !SetTransferMode( gCurrPort, ECP_TRANSFER ) )
					{
 					fprintf(stderr, "TransferMsg: SetTransferMode failed\n" );
					goto TransferError;
					}
				if ( !Forward2Reverse( gCurrPort ) )
					{
					EndTransferMode( gCurrPort );
					goto TransferError;
					}
				if ( gCurrPort->PortType & ECP_PORT_TYPE )
					{
					if ( !ECPReadBuffer(\
					gCurrPort, DataPtr, (DWORD)PACKET_LEN ) )
						{
						fprintf(stderr, "TransferMsg: ECPReadBuffer failed\n" );
						goto TransferError;
						}
					}
				else
					{
					if ( !SimECPReadBuffer(\
					gCurrPort, DataPtr, (DWORD)PACKET_LEN ) )
						{
						fprintf(stderr, "TransferMsg: SimECPReadBuffer failed\n" );
						goto TransferError;
						}
					}
				if( !EndTransferMode( gCurrPort ) )
					{
					fprintf(stderr, "TransferMsg: EndTransferMode failed\n" );
						goto TransferError;
					}
				}/* end if BIDIR_PORT_TYPE */
			else /* not BIDIR_PORT_TYPE */
				{
				if ( !SetTransferMode( gCurrPort, NIBBLE_TRANSFER ) )
					{
					fprintf(stderr, "TransferMsg: SetTransferMode failed\n" );
					goto TransferError;
					}
				if ( !NibbleReadBuffer(\
				gCurrPort, DataPtr, (DWORD)PACKET_LEN ) )
					{
					fprintf(stderr, "TransferMsg: NibbleReadBuffer failed\n" );
					goto TransferError;
					}
				if( !EndTransferMode( gCurrPort ) )
					{
					fprintf(stderr, "TransferMsg: EndTransferMode failed\n" );
					goto TransferError;
					}
				}/* end if not BIDIR_PORT_TYPE */
			} /* end if DataReadFlag */
		gDataPacketsPending--;
		DataPtr += PACKET_LEN; 
		}/* end while gDataPacketsPending */
	}/* end if dataBytes != 0 */
	
return( TRUE );

TransferError:
	return( FALSE );
}/* end function CLINK_TransferMsg */



BOOL CLINK_UploadStreamData(
BOOL EndOnFF, DWORD DataBytes, BYTE  __huge *DataBuffer,\
DWORD *ActualBytes, DWORD StreamChunkSize )
{
DWORD curr;

if ( gCurrPort == NULL ) return( FALSE );

//fprintf(stderr, "XX CLINK_UploadStreamData(): porttype=%d\n", gCurrPort->PortType);

if ( gCurrPort->PortType & BIDIR_PORT_TYPE )
	{
//fprintf(stderr, "XX BIDIR\n");
	if ( !SetUploadMode( gCurrPort, ECP_UPLOAD, NORMAL_DATA_FLAG ) )
/*	if ( !SetUploadMode( gCurrPort, ECP_UPLOAD, RLE_DATA_FLAG ) )*/
		{
		fprintf(stderr, "UploadStreamData: SetUploadMode failed\n" );
		goto UploadError;
		}
	if ( !EndOnFF )
		{
		if ( gCurrPort->PortType & ECP_PORT_TYPE )
			{
			if ( !ECPReadBuffer( gCurrPort, DataBuffer, DataBytes ) )
				{
				fprintf(stderr, "UploadStreamData: ECPReadBuffer failed\n" );
				EndUploadMode( gCurrPort );
				goto UploadError;
				}
			}
		else
			{
			if ( !SimECPReadBuffer( gCurrPort, DataBuffer, DataBytes ) )
				{
				fprintf(stderr, "UploadStreamData: SimECPReadBuffer failed\n" );
				EndUploadMode( gCurrPort );
				goto UploadError;
				}
			}
		*ActualBytes = DataBytes;
		}
	else
		{
		if ( (StreamChunkSize < 256) || (StreamChunkSize > DataBytes) 
		|| ( (DataBytes % StreamChunkSize) != 0 ) ) return( FALSE );
			
		for( curr = 0; curr < DataBytes; curr += StreamChunkSize )
			{
			if ( gCurrPort->PortType & ECP_PORT_TYPE )
				{
				if ( !ECPReadBuffer(\
				gCurrPort, &DataBuffer[ curr ], StreamChunkSize ) )
					{
					fprintf(stderr, "UploadStreamData: ECPReadBuffer failed\n" );
					EndUploadMode( gCurrPort );
					goto UploadError;
					}
				}
			else
				{
				if ( !SimECPReadBuffer(\
				gCurrPort, &DataBuffer[ curr ], StreamChunkSize ) )
					{
					fprintf(stderr, "UploadStreamData: SimECPReadBuffer failed\n" );
					EndUploadMode( gCurrPort );
					goto UploadError;
					}
				}

			*ActualBytes = curr + StreamChunkSize;

			if ( (DataBuffer[curr+StreamChunkSize-4] == 0xFF)
				&& (DataBuffer[curr+StreamChunkSize-3] == 0xFF) 
				&& (DataBuffer[curr+StreamChunkSize-2] == 0xFF)
				&& (DataBuffer[curr+StreamChunkSize-1] == 0xFF) ) break;
			}
		}
	if( !EndUploadMode( gCurrPort ) )
		{
		fprintf(stderr, "UploadStreamData: EndUploadMode failed\n" );
		goto UploadError;
		}
	}
else
	{
//fprintf(stderr, "XX NIBBLE\n");	
	if ( !SetUploadMode( gCurrPort, NIBBLE_UPLOAD, NORMAL_DATA_FLAG ) )
/*	if ( !SetUploadMode( gCurrPort, NIBBLE_UPLOAD, RLE_DATA_FLAG ) )*/
		{
		fprintf(stderr, "UploadStreamData: SetUploadMode failed\n" );
		goto UploadError;
		}
	if ( !EndOnFF )
		{
		if ( !NibbleUpload( gCurrPort, DataBuffer, DataBytes ) )
			{
			fprintf(stderr, "UploadStreamData: NibbleUpload failed\n" );
			EndUploadMode( gCurrPort );
			goto UploadError;
			}
		*ActualBytes = DataBytes;
		}
  	else
  		{
		for( curr = 0; curr < DataBytes; curr += StreamChunkSize )
			{
			if ( !NibbleUpload(\
			gCurrPort, &DataBuffer[curr], StreamChunkSize ) )
				{
				fprintf(stderr, "UploadStreamData: NibbleUpload failed\n" );
				EndUploadMode( gCurrPort );
				goto UploadError;
				}

			*ActualBytes = curr + StreamChunkSize;

			if ( (DataBuffer[curr+StreamChunkSize-4] == 0xFF)
			&& (DataBuffer[curr+StreamChunkSize-3] == 0xFF) 
			&& (DataBuffer[curr+StreamChunkSize-2] == 0xFF)
			&& (DataBuffer[curr+StreamChunkSize-1] == 0xFF) ) break;
			}
		}
  	if( !EndUploadMode( gCurrPort ) )
  		{
		fprintf(stderr, "UploadStreamData: EndUploadMode failed\n" );
		goto UploadError;
		}
	}

return( TRUE );

UploadError:
	return( FALSE );
}
