#ifndef _INTERNET_H_
#define _INTERNET_H_

/*#include <termcap.h>*/

#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/file.h>
#include <sys/socket.h>

#include <pwd.h>

#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>

#include <errno.h>
#include <unistd.h>

/*
maximum allowed connections, not to many, now >200kB per picture per client!
*/
/*
This does NOT mean only one connection, just the wait queue size in listen()
*/
#define MAX_LISTEN_QUEUE 1

/* commands to server */
#define HIGH_POWER						1 
#define LOW_POWER						2 
#define RESET							3 
#define COMPRESSION_ON					4
#define COMPRESSION_OFF					5 
#define BIG_PICURE						6 
#define SMALL_PICURE					7 
#define AUTO_EXPOSURE					8 
#define AUTO_WHITE_BALANCE				9 
#define CONTRAST						10 
#define BRIGHTNESS						11 
#define SATURATION						12 
#define MAN_WHITE_BALANCE				13 
#define MAN_EXPOSURE					14
#define GAIN_CEILING					15
#define CAMERA_RESTART					16
#define COMPENSATION_GAIN				17
#define NO_CODING						18
#define JPEG_CODING						19
#define MPEG2_CODING					20
#define WAVELET_CODING					21
#define CODING_QUALITY					22
#define CLOCK_ON						23
#define CLOCK_OFF						24
#define CLOCK_POSITION					25
#define CLOCK_UTC						26
#define CLOCK_LOCAL						27
#define TEXT_TO_SCREEN					28
#define TEXT_TO_CONSOLE					29
#define TEXT_POSITION					30
#define TEXT_COLOR						31
#define CHAR_SET_S14					32
#define CHAR_SET_TXT					33
#define SET_APERTURE_CORRECTION			34
#define SET_APCOR_DEFAULT				35
#define SET_SENSOR_FRAMES_PER_SECOND	36

/* offsets in txbuffer header for server and client */
#define START_PASSWORD				0	/* send by client only */
#define DATA_SIZE_VERY_HIGH			0	/* send by server only */
#define DATA_SIZE_HIGH				1	/* ,, */
#define DATA_SIZE_LOW				2	/* ,, */
#define TEXT_MSG_NUMBER_HIGH		3	/* ,, */
#define TEXT_MSG_NUMBER_LOW			4	/* ,, */
#define START_SERVER_TEXT			5	/* ,, */
#define START_CLIENT_TEXT			16	/* send by client only */
#define SERVER_STATUS				85	/* send by server only */
#define START_COMMAND				119	/* send by client only */
#define FLAGS						121	/* send by server, part of status */
#define FLAGS2						126 /* send by server, part of server status */
#define HEADER_SIZE					128 /* size of the header */

/* Note: I always transmit high byte first */
/*
structure of the client transmit buffer:
 START_PASSWORD			password (max 16 characters)
 START_CLIENT_TEXT		text (upto START_COMMAND)
 START_COMMAND			command number
 START_COMMAND + 1		command arg 1
 START_COMMAND + 2		command arg 2
 START_COMMAND + 3		command arg 3
 START_COMMAND + 4		command arg 4
 START_COMMAND + 5		command arg 5
 START_COMMAND + 6		command arg 6
 START_COMMAND + 7		command arg 7
 START_COMMAND + 8		command arg 8

structure of the server transmit buffer:
 DATA_SIZE_VERY_HIGH		frame size highest byte		camera download size
 DATA_SIZE_HIGH				frame size high byte		,,
 DATA_SIZE_LOW				frame size low byte			,,
 TEXT_MSG_NUMBER_HIGH		text message number high byte
 TEXT_MSG_NUMBER_LOW		text message number low byte 
 START_SERVER_TEXT			text (upto SERVER_STATUS)
 SERVER_STATUS + 0			apertur correction treshhold 1 intensity 1 
 SERVER_STATUS + 1			apertur correction treshhold 2 intensity 2
 SERVER_STATUS + 2			apertur correction treshhold 3 intensity 3
 SERVER_STATUS + 3			apertur correction treshhold 4 intensity 4
 SERVER_STATUS + 4			server version, high nibble major, low nibble minor
 SERVER_STATUS + 5
 SERVER_STATUS + 6
 SERVER_STATUS + 7
 SERVER_STATUS + 8
 SERVER_STATUS + 9
 SERVER_STATUS + 10			server_frames_per_second
 SERVER_STATUS + 11			brightness	status server starts here
 SERVER_STATUS + 12			contrast
 SERVER_STATUS + 13 		saturation	
 SERVER_STATUS + 14			current red gain
 SERVER_STATUS + 15			current green gain
 SERVER_STATUS + 16			current blue gain
 SERVER_STATUS + 17			current gain (returns gain_ceiling - 1)
 SERVER_STATUS + 18			current fine exposure
 SERVER_STATUS + 19			current coarse exposure high byte
 SERVER_STATUS + 20			current coarse exposure low byte
 SERVER_STATUS + 21		red compensation
 SERVER_STATUS + 22		green 1 compensation
 SERVER_STATUS + 23		green 2 compensation
 SERVER_STATUS + 24		blue compensation
 SERVER_STATUS + 25		text x low byte	
 SERVER_STATUS + 26		text y low byte
 SERVER_STATUS + 27		text fr
 SERVER_STATUS + 28		text fg
 SERVER_STATUS + 29		text fb
 SERVER_STATUS + 30		text br
 SERVER_STATUS + 31		text bg
 SERVER_STATUS + 32		text bb
 SERVER_STATUS + 33		coding quality (0 - 100)
 SERVER_STATUS + 34		clock position
 SERVER_STATUS + 35		number of connected clients
 SERVER_STATUS + 36		flags:	feedback server to client
 									bit 0 you have command permission
								 	bit 1 bit 0 coding feedback 
									bit 2 0=bytes/sec 1=kB/sec in +28 +29 
									bit 3 bit 1 coding feedback
									bit 4 man exposure
								 	bit 5 man white
								 	bit 6 clock on
								 	bit 7 clock 1 = UTC / 0 = local
 SERVER_STATUS + 37		frames per thousand sec high
 SERVER_STATUS + 38		frames per thousand sec low 
 SERVER_STATUS + 39		bytes per sec high
 SERVER_STATUS + 40		bytes per sec low
 SERVER_STATUS + 41		flags2: feedback server to client
 									bit 0 1=big picture 0=small picture 
 									bit 1 show_text_flag
 									bit 2 text x bit 8
 									bit 3 text y bit 8
 									bit 4 0=s14font, 1=txtfont 
 									5-7 reserved.
 SERVER_STATUS + 42		reserved	96 + 31 = 127 is last usable byte 
 HEADER_SIZE				the data downloaded from the camera starts here,
								with its own 64 byte header.
*/

/* prototypes */
extern int make_socket (unsigned short int port);
extern int accept_connections(\
	int argc, char **argv, int port, int coding_mode);
extern int init_sockaddr (\
	struct sockaddr_in *name, const char *hostname, unsigned short int port);
extern int read_from_server(\
	int argc, char **argv, char *server, int port, char *password);
extern int construct_camera_command(char *text, BYTE __huge *buffer);
extern int execute_camera_command(BYTE __huge *buffer, int *coding_mode);
extern int show_commands();
extern int restart_camera();
extern int camera_status_to_buffer(BYTE __huge *buffer);
extern int show_status(BYTE __huge *buffer);
extern int execute_local_commands(char *text);
extern int server_status_to_buffer(\
	BYTE __huge *buffer, int clients,\
	float bytes_per_second, float frames_per_second);

#endif /* _INTERNET_H_ */

